!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
!! Wavefield propagation module
!! MOdified from propdecom_mod
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 

module propdecomp_mod
  use sep
  use fftshift_mod
  use fkfilter_mod
  implicit none
  include "fftw3.f"

  logical       :: wfldout                                 ! Switch for output wavefield movie
  real, private :: dt                                      ! Propagation time step size
  integer,private :: nt, jt, jst                               ! Ratio between dt and dt_data
  integer,private :: itdelay                               ! itdelay
  integer,private :: nt_data, nx, nz, nnx, nnz             ! Size of data
  integer,private :: nop, npad                             ! order of Laplacian operator
  real, private,pointer :: v(:,:), v2(:,:)                          ! Velocity 
  real*8, private         :: a1x, a2x, a3x, a4x, a0          ! FD coef
  real*8, private         :: a1z, a2z, a3z, a4z              ! FD coef
  real, private, pointer :: src_func(:,:),rec_func(:,:)    ! src wavelet, receiver data
  integer,private, pointer  :: src_loc(:,:),rec_loc(:,:)   ! source and receiver location
  real, private, pointer :: mult(:,:)                    ! For absorbing boundary
  real, private, pointer :: zxtaper(:),ctaper(:,:)       ! Taper
  real                   :: ovel(2),dvel(2)              ! Sampling of velocity model
  character(len=128)              :: wfldsou_, wfldimg_, wfldrec_

  !kz filter related parameter
  real                           :: alpha   ! for kz filtering
  integer                        :: filter
  
  contains

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  
!! SUBROUTINE #1: INITIALIZE PROPAGATION
!! This subroutine is to prepare for propagation by calculating necessary parameters and coeffients
!! 1) Figure out a dt that is an integer ratio of dt_data for propagation and passes stability
!! 2) Calculate coefficients 
!! 3) Set velocity sampling, it will just be the same as in the input velocity model. 
!! 4) Other definitions
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  
  
  subroutine init_propagate(vel,oz,ox,dz,dx,dt_data,nt_use,wfldout_in,wfldsou_in,wfldrec_in,wfldimg_in,vmax,npad_in) 
    logical          :: wfldout_in
    real, target     :: vel(:,:) !Input velocity model
    real             :: dz,dx !Sampling of input field
    real*8             :: idx, idz    
    real             :: dt_data,  val  !Sampling of the data
    real             :: oz,ox   !Origin of velocity model
    real             :: vmax    !the maximum velocity contained in the velocity model
    integer          :: ix,iz,nt_use,k,p, npad_in !integral ratios between dt_data and dt, origins
    character(len=128)              :: wfldsou_in, wfldrec_in, wfldimg_in   
    
    write(0,*) 'Running: init_propagate'

    ! (1) The stability condition to calculate the new dt from dt_data:
 !   vmax=MAXVAL(vel)
    nt_data=nt_use
    if (dx.ge.dz) then
	k=ceiling(dt_data/(0.707*dz/vmax))+1	!using ceiling to get the largest possible integer,
	!since we need vmax*dt to be less than 0.707 (2D Courant number)
	dt=dt_data/k
	jt=k
       ! write(0,*) 'init_', vmax, nt_data, k, dt_data, dt,jt, dz
    else !if dz .gt. dx
	p=ceiling(dt_data/(0.707*dx/vmax))+1	
	dt=dt_data/p
	jt=p
       ! write(0,*) 'init_', vmax, nt_data, k, dt_data, dt,jt, dx
    end if

    nt = nt_data*jt


    !(2) Coefficient calculation (8th order finite diff)
    idx=DBLE(dt*dt/(dx*dx))
    idz=DBLE(dt*dt/(dz*dz))	
    a0 =  DBLE(-2.84722222222)*(idx+idz)
    a1x =  DBLE(1.6000000*idx);       a1z= DBLE(1.6000000*idz)       	
    a2x =  DBLE(-0.2000000*idx);      a2z=DBLE(-0.2000000*idz)       	
    a3x =  DBLE(0.0253968254*idx);    a3z=DBLE(0.0253968254*idz)       	
    a4x =  DBLE(-0.00178571428*idx);  a4z=DBLE(-0.00178571428*idz)       	
 
    !(3) Find the velocity parameters
    v=>vel      
    ovel(1) = oz;    ovel(2) = ox
    dvel(1) = dz;    dvel(2) = dx


   !(4) Other defn
   nop=4;   ! Using 8th order Laplacian
   wfldout=wfldout_in
   wfldsou_ = wfldsou_in
   wfldrec_ = wfldrec_in
   wfldimg_ = wfldimg_in

   
   nx=size(v,2);
   nz=size(v,1);
   nnx=nx-2*npad
   nnz=nz-2*npad

   !(5) Setup damping boundary
   npad=npad_in
   call init_taperCos(npad)
   
   !(6) Velocity padding 
   do iz=1,npad
      v(iz,:)    = v(npad+1,:)
      v(nz-iz,:) = v(nz-npad-1,:)
      v(:,iz)    = v(:,npad+1)
      v(:,nx-iz) = v(:,nx-npad-1)
   end do
  
   ! (7) square the velocity
   allocate(v2(nz,nx))
   do ix=1,nx
      do iz=1,nz
         v2(iz,ix) = v(iz,ix)*v(iz,ix)
      end do
   end do

      
   !(8) Define one-way boundary
   call define_mult()

  end subroutine
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! SUBROUTINE #2: INITIALIZE THE SOURCE
!! This subroutine define the parameters necessary to propagate source wavefields 
!! 1) Allocate src_loc(2,1): the size of this array is 2 rows by 1 column
!!    Allocate src_func(nt,1) - store the source-wavelet in time
!! 2) Set src_loc location
!! 3) Interpolate source to src_func to fit the number of steps we want to propagate
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  
  subroutine source_init(source,xsrc,zsrc,tdelay,dst)
    real, allocatable, dimension(:) :: source !Source function 
    real :: xsrc,zsrc !Source location in initial velocity model coordinate system !!
    integer :: i1,i0 !number of sources
    real :: fract, tdelay, dst


    write(0,*) 'Running: source_init'

    !(0) Itdelay
    itdelay = (tdelay)/dt + 0.5
    jst = dst/dt

    write(0,*) 'jst, jt', jst, jt
    write(0,*) 'dt, dst', dt, dst

    !(1) Allocations:
    allocate(src_loc(2,1), src_func(nt+itdelay,1))

    !(2) Set src_loc
    src_loc(1,1)=(zsrc-ovel(1))/dvel(1)+1.5 + max(nop,npad)
    src_loc(2,1)=(xsrc-ovel(2))/dvel(2)+1.5 + max(nop,npad)
    ! error check to make sure inside grid
    if ((src_loc(1,1) .gt. nnz+npad) .and. (src_loc(1,1) .lt. 1)) call erexit('src z location out of bound')
    if ((src_loc(2,1) .gt. nnx+npad) .and. (src_loc(2,1) .lt. 1)) call erexit('src x location out of bound')

    !(3) Interpolate source to src_func to fit the number of steps in propagation   
    src_func=0
    ! linear interpolation
    do i0=1,size(source)
       do i1=1,jst
          fract=real(i1-1)/real(jst)
          src_func(i1+(i0-1)*jst,1)=(1.-fract)*source(i0)+fract*source(min(i0+1,size(source)))
       end do
    end do

!   ! Take a second Derivative of the source
!   call FirstDer(.true.,src_func(:,1),src_func(:,1),size(src_func,1),1.)

  end subroutine
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!    
!! SUBROUTINE #3: RECEIVER INITIALIZATION 
!! Initialize the parameter necessary for receivers wavefield or location
!! 1) Allocate rec_loc(2,size(recs,2)) rec_func(size(recs)*jt,size(recs,2))
!! 2) Set rec_loc locations
!! 3) Interpolate recs to rec_func
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

 subroutine receiver_init(direction,zsrc,xsrc,xoff_o,xoff_d,recs)
   logical :: direction   ! [0] modeling [1] migration
   real    :: zsrc, xsrc  !Position of the source
   real    :: xoff_o,xoff_d !Sampling of the receivers compared to the source position
   real, allocatable, dimension(:,:)    :: recs !Recorded data
   integer :: ir,irx,io,il,iit
   real    :: fract
   
   
   write(0,*) 'Running: receiver_init'
   ! write(0,*) 'xsrc, zsrc', xsrc, zsrc
   
   !(1) Allocation
   allocate (rec_loc(2,size(recs,2)), rec_func(size(recs,1)*jt,size(recs,2)))   

   !(2) Set rec_loc 
   do irx=1,size(recs,2)
     rec_loc(2,irx)=(xoff_o+irx*xoff_d-ovel(2))/dvel(2)+0.5 + max(nop,npad)
   end do
   rec_loc(1,:)=(zsrc-ovel(1))/dvel(1)+0.5 + max(nop,npad)
   !check to make sure inside grid
   write(0,*) 'rec z', rec_loc(1,1), zsrc, nnz, npad
   if ( (rec_loc(1,1) .lt. 1) .or. (rec_loc(1,1) .gt. nnz+npad)) call erexit('receiver z location out of bound')
   do irx=1,size(recs,2)
      ir = rec_loc(2,irx)
      if ( (ir .lt. 1) .or. (ir .gt. nnx+npad)) call erexit('receiver x location out of bound')
   end do



  
   !(3) Interpolate the data in time from sampling dt_data to dt (store in reverse time)
   rec_func=0.
   if (direction) then
      ! write(0,*) 'interpolate rec func'
      do irx=1,size(recs,2)
         do io=size(recs,1),1,-1
            do il=1,jt
               fract=real(il-1)/real(jt)
               iit = nt_data*jt-1*(-il+io*jt+1)
               rec_func(iit,irx)=(1.-fract)*recs(io,irx)+fract*recs(min(io+1,size(recs)),irx)
            end do
         end do
      end do
   end if
   
   

 end subroutine
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
!! SUBROUTINE #4: WAVE-FIELD PROPAGATION
!! Input: prev,cur,next =  u(t+1), u(t), u(t-1)
!! Input: rec_f = time-index marker on where to inject data
!! Input: nt = number of time step to propagate
!! output: next (represent the wavefield after propagating nt time steps)
!! (1) Apply Laplacian and time-marching (in reverse time)
!!         u(t+/-1) = - u(t-/+1) + 2*u(t) + (v^2)*(dt^2)*(Laplacian)u(t)
!! Note that propagate forward and reverse in time is symmetric
!! (2) data/source injection       
!! (3) Time stepping (update the 3 time slices)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

 subroutine propagate_wavefield(prev,cur,next,wave_func,wave_loc,time_marker,nt)
    real, pointer, dimension(:,:)    :: prev,cur,next !Previous,current, and next wavefields
    real, pointer                    :: wave_func(:,:)
    integer, pointer                 :: wave_loc(:,:)
    integer :: time_marker             !Current time step marker for injection
    integer :: ix,iz,it,nt,ir,irx
    real*8 :: val

   !write(0,*) 'Running: propagate_wavefield', time_marker ,'out of', nt_data*jt, sum(prev),sum(cur),sum(next)

  if (mod(time_marker,50) == 0) write(0,*) 'Running: propagate_wavefield', time_marker ,'out of', nt_data*jt , sum(prev), sum(cur), sum(next)

    do it=1,nt
       val = 0.
        !(1) source/receiver injection
        do ir=1,size(wave_loc,2)
           cur(wave_loc(1,ir),wave_loc(2,ir))=cur(wave_loc(1,ir),wave_loc(2,ir))+wave_func(it+time_marker,ir)
        end do       
       
       ! (1) Apply Laplacian and time-marching
       do ix=1+nop,nx-nop   
          do iz=nop+1,nz-nop
             val = a0*cur(iz,ix)
             val = val + a1x*DBLE(cur(iz,ix+1))
	     val = val + a1x*DBLE(cur(iz,ix-1))
             val = val + a2x*DBLE(cur(iz,ix+2))
	     val = val + a2x*DBLE(cur(iz,ix-2))
             val = val + a3x*DBLE(cur(iz,ix+3))
	     val = val + a3x*DBLE(cur(iz,ix-3))
             val = val + a4x*DBLE(cur(iz,ix+4))
	     val = val + a4x*DBLE(cur(iz,ix-4))
             val = val + a1z*DBLE(cur(iz+1,ix))
	     val = val + a1z*DBLE(cur(iz-1,ix))
             val = val + a2z*DBLE(cur(iz+2,ix))
	     val = val + a2z*DBLE(cur(iz-2,ix))
             val = val + a3z*DBLE(cur(iz+3,ix))
	     val = val + a3z*DBLE(cur(iz-3,ix))
             val = val + a4z*DBLE(cur(iz+4,ix))
	     val = val + a4z*DBLE(cur(iz-4,ix))
 	     val = val*DBLE(v2(iz,ix))

	     val = val + DBLE(-prev(iz,ix) + 2*cur(iz,ix))
	     next(iz,ix) = REAL(val)
           end do
        end do

        call update_sponge(prev,cur,next)

!        ! Applying damping boundary
!        call taperingCos(prev,cur,next)

        !(3) Time stepping
        prev=cur
        cur=next
        
       call update_oneway(prev,cur,next)
        
   end do   
end subroutine

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
!! SUBROUTINE #5: WAVE-FIELD PROPAGATION Adjoint
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

 subroutine propagate_wavefield_adj(prev,cur,next,time_marker,nt)
    real, pointer, dimension(:,:)    :: prev,cur,next !Previous,current, and next wavefields
    integer :: time_marker             !Current time step marker for injection
    integer :: ix,iz,it,nt
    real :: val

! write(0,*) 'Running: propagate_wavefield', time_marker ,'out of', nt_data*jt

  if (mod(time_marker,50) == 0) write(0,*) 'Running: propagate_wavefield_adj', time_marker ,'out of', nt_data*jt

    do it=1,nt
       val = 0.
       
       ! (1) Apply Laplacian and time-marching
       do ix=1+nop,nx-nop   
          do iz=nop+1,nz-nop


             val = a0*v2(iz,ix)*cur(iz,ix)
             val = val + a1x*v2(iz,ix+1)*cur(iz,ix+1)
             val = val + a1x*v2(iz,ix-1)*cur(iz,ix-1)
             val = val + a2x*v2(iz,ix+2)*cur(iz,ix+2)
             val = val + a2x*v2(iz,ix-2)*cur(iz,ix-2)
             val = val + a3x*v2(iz,ix+3)*cur(iz,ix+3)
             val = val + a3x*v2(iz,ix-3)*cur(iz,ix-3)
             val = val + a4x*v2(iz,ix+4)*cur(iz,ix+4)
             val = val + a4x*v2(iz,ix-4)*cur(iz,ix-4)
             val = val + a1z*v2(iz+1,ix)*cur(iz+1,ix)
             val = val + a1z*v2(iz-1,ix)*cur(iz-1,ix)
             val = val + a2z*v2(iz+2,ix)*cur(iz+2,ix)
             val = val + a2z*v2(iz-2,ix)*cur(iz-2,ix)
             val = val + a3z*v2(iz+3,ix)*cur(iz+3,ix)
             val = val + a3z*v2(iz-3,ix)*cur(iz-3,ix)
             val = val + a4z*v2(iz+4,ix)*cur(iz+4,ix)
             val = val + a4z*v2(iz-4,ix)*cur(iz-4,ix)	     
	     val = val + -prev(iz,ix) + 2.*cur(iz,ix)
	     next(iz,ix) = val

           end do
        end do

        call update_sponge(prev,cur,next)

        !(3) Time stepping
        prev=cur
        cur=next
        
        call update_oneway(prev,cur,next)
        
   end do   
end subroutine


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
!! SUBROUTINE #6: IMAGING
!! (1) apply zero-lag cross correlation imaging condition
!! (2) Output to wavefield movie (if optioned)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 

 subroutine imaging_condition(source0,source1,receiver,image,io,jt)
    real, dimension(:,:)      :: image
    real, dimension(:,:)   ::  source0,source1, receiver
    integer                   :: ix, iz, irx,io,jt

    !(1) Apply imaging condition
    do ix=1,nx
       do iz=1,nz
          image(iz,ix) = image(iz,ix) + real((source0(iz,ix)*(jt-io)+io*source1(iz,ix))*receiver(iz,ix))
       end do
    end do

    !(2) Output wavefield movies  
!    if (wfldout) then
!       call sep_write(image,wfldimg_)
!       call sep_write(source,wfldsou_)
!       call sep_write(receiver,wfldrec_)
!    end if

 end subroutine

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
!! SUBROUTINE #7: IMAGING
!! (1) apply zero-lag cross correlation imaging condition
!! (2) Output to wavefield movie (if optioned)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 

 subroutine imaging_conDecomp(source0,source1, receiver,source2,source3,receiver2,image,io,jt)
    real, dimension(:,:)      :: image
    complex, dimension(:,:)   ::  source0,source1, receiver
    complex, dimension(:,:)   ::  source2, source3, receiver2
    integer                   :: ix, iz, irx, io,jt

    !(1) Apply imaging condition
    do ix=1,nx
       do iz=1,nz
          image(iz,ix) = image(iz,ix) + real((source0(iz,ix)*(jt-io)+io*source1(iz,ix))*receiver(iz,ix)) 
          image(iz,ix) = image(iz,ix) + real((source2(iz,ix)*(jt-io)+io*source3(iz,ix))*receiver2(iz,ix)) 
       end do
    end do

 
 end subroutine
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
!! SUBROUTINE #7: IMAGING
!! (1) apply zero-lag cross correlation imaging condition
!! (2) Output to wavefield movie (if optioned)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 

 subroutine imaging_conDecompV2(source0,source1, receiver,source2,source3,receiver2,image,io,jt)
    real, dimension(:,:)      :: image
    complex, dimension(:,:)   ::  source0,source1, receiver
    complex, dimension(:,:)   ::  source2, source3, receiver2
    integer                   :: ix, iz, irx, io,jt

    !(1) Apply imaging condition
    do ix=1,nx
       do iz=1,nz
          image(iz,ix) = image(iz,ix) + real((source0(iz,ix)*(jt-io)+io*source1(iz,ix))*receiver(iz,ix)) 
          image(iz,ix) = image(iz,ix) + real((source2(iz,ix)*(jt-io)+io*source3(iz,ix))*receiver2(iz,ix)) 
       end do
    end do

 
    !(2) Output wavefield movies  
    if ((wfldout) .and. (io .eq. 0)) then
       call sep_write(image,wfldimg_)
       call sep_write(source0,wfldsou_)
       call sep_write(receiver,wfldrec_)
    end if

 end subroutine

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
!! SUBROUTINE #8: Forward Modeling Born Injection
!! (1) Apply Born source injection to receiver wavefield
!! (2) Output wavefield movies  
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
 subroutine forward_born_inject(source0,source1,receiver,image,io,jt)

    real, dimension(:,:) :: image, receiver
    complex, dimension(:,:) :: source0, source1
    integer              :: ix, iz,io,jt

   ! write(0,*) 'forward_born_inject', io, jt

    !(1) Apply Born source injection to receiver wavefield
    do ix=1,nx
       do iz=1,nz
          receiver(iz,ix) = receiver(iz,ix) + real((source0(iz,ix)*(jt-io)+source1(iz,ix)*io))*image(iz,ix)
       end do
    end do   

    !(2) Output wavefield movies  
    if ((wfldout) .and. (io.eq.0)) then
       call sep_write(image,wfldimg_)
       call sep_write(source0,wfldsou_)
       call sep_write(receiver,wfldrec_)
    end if

 end subroutine

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! SUBROUTINE #9: Propagate and save source wavefield
!! (1) Allocation
!! (2) propagate source wavefield
!! (3) Save to array at every jt time-step
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 

  subroutine create_source_array(array,arrUp,arrDn)
  real, allocatable, dimension(:,:,:)  :: array !Array of nz,nx, nt_data
  complex, allocatable, dimension(:,:,:)  :: arrUp, arrDn !Array of nz,nx
  integer :: i, irx                    
  real, pointer, dimension(:,:) :: prev,cur,next

  write(0,*) 'Running: create_source_array', itdelay, dt, dt*jt, jt
  ! (1) Allocation  

  allocate(prev(nz,nx),cur(nz,nx),next(nz,nx))


  do i=1,nt_data+itdelay/jt
     ! (2) Propagate source wavefield for jt time-steps
     call propagate_wavefield(prev,cur,next,src_func,src_loc,(i-1)*jt+1,jt)
     ! (3) Save a time-slice of sou wavefield to the array at every jt time-step
     if (i .gt. itdelay/jt) then
!        write(0,*) 'i, it_cur, size(array,3)', i, (i-1)*jt+1, size(array,3), itdelay
        array(:,:,i-itdelay/jt)=next
        ! nextup, next down
        call decompNext1D(next,arrUp(:,:,i-itdelay/jt),arrDn(:,:,i-itdelay/jt))
     end if
  end do
  
  

  
  deallocate(prev,cur,next)

end subroutine

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! SUBROUTINE #10: Decompose a 2d slice by kz filtering
!! assume 1-axis is z
!! filter=[0]do nothing; [1]sharp cut; [2]Tukey filter using alpha parameter
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
subroutine decompNext1D(next,Up,Dn)

   integer                       :: n1, n2, i, j, stat, flag
   real, pointer, dimension(:,:) :: next
   complex, dimension(:,:)          :: Up, Dn
   complex, dimension(:,:), allocatable  :: in, in1, in2
   complex, dimension(:), allocatable    :: out
   real, dimension(:), allocatable    :: temp

   integer*8 plan1f
   alpha = 0.3
   filter = 1;

   n1 = size(next,1)
   n2 = size(next,2)
   allocate(in(n1,n2), out(n1), temp(n1))
   allocate(in1(n1,n2), in2(n1,n2))  


   do i=1,n2
      do j=1,n1
         in(j,i) = cmplx(next(j,i),0)
	 Up(j,i) = 0.;
	 Dn(j,i) = 0.;
      end do
   end do


   !! Forward FFT
   call sfftw_plan_dft_1d(plan1f, n1, in(:,1), out, FFTW_BACKWARD, FFTW_ESTIMATE)
   do i=1,n2
      call sfftw_execute_dft(plan1f, in(:,i), out(1:n1))
      in(:,i)=out(1:n1)/sqrt(real(n1)) ! divided by sqrt(n1) instead of n1. this helps when computing correlation in FFT domain
   end do
   call sfftw_destroy_plan(plan1f)



   !! Do shifting from [0,2*fnyq) to [-fnyq,fnyq) so that we can separate wavefield components
   do i=1,n2
     stat=fftshift1Dc_lop(.false.,.false.,in(:,i),out) !fftshift1Dc_lop(adj,add,x,y)
     in(:,i) = out(1:n1)
   end do


   !! Filtering
   ! storing negative axis
   flag = 1
   call filter1Dfn(temp(:),filter,flag,alpha=alpha) ! create the temp function corresponding to the selected filter and flag.
   do i=1,n1
      in1(i,:) = in(i,:) * temp(i)
   end do

   !! Storing positive axis
   flag = -1
   call filter1Dfn(temp(:),filter,flag,alpha=alpha) ! create the temp function corresponding to the selected filter and flag.
   do i=1,n1
      in2(i,:) = in(i,:) * temp(i)
   end do
   deallocate(temp)


   !! Do reshifting from [-fnyq,fnyq) to [0,2*fnyq) so that they can enter iFFT program
   do i=1,n2
      stat=fftshift1Dc_lop(.true.,.false.,out,in1(:,i)) !fftshift1Dc_lop(adj,add,x,y)
      do j=1,n1
       in1(j,i)=out(j)
      end do
   end do
   do i=1,n2
      stat=fftshift1Dc_lop(.true.,.false.,out,in2(:,i)) !fftshift1Dc_lop(adj,add,x,y)
      do j=1,n1
      in2(j,i)=out(j)
      end do
   end do



   !! Inverse FFT
   call sfftw_plan_dft_1d(plan1f, n1, in(:,1), out(1:n1), FFTW_FORWARD, FFTW_ESTIMATE)
   do i=1,n2
      call sfftw_execute_dft(plan1f, in1(:,i), out(1:n1))
      do j=1,n1
         Up(j,i)=out(j)/sqrt(real(n1)) ! divided by sqrt(n1) instead of 1
      end do
      call sfftw_execute_dft(plan1f, in2(:,i), out(1:n1))
      do j=1,n1
         Dn(j,i)=out(j)/sqrt(real(n1)) ! divided by sqrt(n1) instead of 1
      end do
   end do
   call sfftw_destroy_plan(plan1f)
   
!   do i=1,n2
!      do j=1,n1
!        Up(j,i) = aimag(in1(j,i))
!        Dn(j,i) = aimag(in2(j,i))
!      end do
!   end do


end subroutine

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! SUBROUTINE #11: Propagate receiver wavefield and apply imaging condition
!! (1) Allocation
!! (2) propagate receiver wavefield
!! (3) apply imaging condition
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

subroutine propagate_receiversdecomp(imageC,imageUUDD,imageUDDU,source_array, souUp, souDn)
  real, dimension(:,:,:)            :: source_array
  complex, dimension(:,:,:)         :: souUp, souDn  
  real, dimension(:,:)              :: imageC
  real, dimension(:,:)              :: imageUUDD, imageUDDU
  integer                           :: i,irx, id, io  !Internal time variable
  real, pointer,dimension(:,:)      :: prev,cur,next
  complex, pointer, dimension(:,:)  :: recUp, recDn

  write(0,*) 'Running: propagate_receivers'  

  ! (1) Allocation
  allocate(prev(nz,nx),cur(nz,nx),next(nz,nx),recUp(nz,nx),recDn(nz,nx))
  prev=0; cur=0; next=0;

  do i=1,nt_data*jt
     ! (2) propagate rec wavefield for jt time-step (note: it is reverse in time)
     call propagate_wavefield(prev,cur,next,rec_func,rec_loc,i,1)
     call decompNext1D(prev,recUp,recDn)
     id=(i-1)/jt+1
     io = i-((id-1)*jt+1)
     if (id .lt. nt_data) then
     ! (3) Apply imaging condition at every jt time-step
     call imaging_condition(source_array(:,:,nt_data-id+1),source_array(:,:,nt_data-id),prev,imageC,io,jt)
     call imaging_condecompV2(souUp(:,:,nt_data-id+1),souUp(:,:,nt_data-id),recUp,souDn(:,:,nt_data-id+1),souDn(:,:,nt_data-id),recDn,imageUUDD,io,jt)
     call imaging_condecomp(souUp(:,:,nt_data-id+1),souUp(:,:,nt_data-id),recDn,souDn(:,:,nt_data-id+1),souDn(:,:,nt_data-id),recUp,imageUDDU,io,jt)
     end if
  end do

  deallocate(prev,cur,next)
  
end subroutine

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! SUBROUTINE #12: Propagate the Born receiver wavefield
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

subroutine propagate_borndecomp(image,dat,source_array,souUp,souDn)
  real, dimension(:,:,:) :: source_array
  complex, dimension(:,:,:)  :: souUp,souDn  
  real, dimension(:,:) :: image, dat
  integer :: i,ix, it, iz, irx,io,id  !Internal time variable
  real, pointer,dimension(:,:) :: prev2,cur2,next2
  real, pointer,dimension(:,:) :: prev3,cur3,next3
  complex, pointer, dimension(:,:)  :: kUp,kDn,dummy

  write(0,*) 'Running: propagate_borndecomp'  

  ! (1) Allocation
  allocate(prev2(nz,nx),cur2(nz,nx),next2(nz,nx))
  allocate(prev3(nz,nx),cur3(nz,nx),next3(nz,nx))
  allocate(kUp(nz,nx),kDn(nz,nx),dummy(nz,nx))

  ! (2) Initialization
  do ix=1,nx
     do iz=1,nz
        prev2(iz,ix)=0; cur2(iz,ix)=0; next2(iz,ix)=0;
        prev3(iz,ix)=0; cur3(iz,ix)=0; next3(iz,ix)=0;
     end do
  end do
  it=0
  
  do i=1,nt_data*jt

     ! (2) propagate rec wavefield 
     call propagate_wavefield_adj(prev2,cur2,next2,i,jt)
     call propagate_wavefield_adj(prev3,cur3,next3,i,jt)

     id = (i-1)/jt+1
     if (id .lt. nt_data) then
	io = i -((id-1)*jt+1)
        ! (3) Inject Born Source at every time-step
        call forward_born_inject(souDn(:,:,id),souDn(:,:,id),cur2,image,io,jt)
        call forward_born_inject(souUp(:,:,id),souUp(:,:,id),cur3,image,io,jt)
     end if


     call decompNext1D(cur2,kUp,dummy)
     call decompNext1D(cur3,dummy,kDn)


     ! (4) Collect data
     if (mod(i,jt)==1) then      
        do irx=1,size(dat,2)
            dat(i,irx) = real(kUp(rec_loc(1,irx),rec_loc(2,irx))+kDn(rec_loc(1,irx),rec_loc(2,irx)))
        end do
     end if

  end do


  deallocate(prev2,cur2,next2)
  deallocate(prev3,cur3,next3)

end subroutine

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! SUBROUTINE #13: MIGRATE A SHOT
!! (1) Allocation
!! (2) Calculate source wavefield
!! (3) Calculate rec wavefield and apply imaging condition
!! (4) Write wavefield movie header (if optioned)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

subroutine migratedecomp_shot(imageC,imageUUDD,imageUDDU)

  real, dimension(:,:) :: imageC ! conventional image
  real, dimension(:,:) :: imageUUDD,imageUDDU  !Decomposed Image
  real, allocatable, dimension(:,:,:)     :: source_array
  complex, allocatable, dimension(:,:,:)  ::souUp, souDn
  
  write(0,*) 'Running: migrateDecomp_shot'

  ! (1) Allocation
  allocate(source_array(nz,nx,nt_data))
  allocate(souUp(nz,nx,nt_data))
  allocate(souDn(nz,nx,nt_data))

  ! (2) Propagate and save source wavefield, kz decomposition
  call create_source_array(source_array,souUp,souDn)

  ! (3) Calculate receiver wavefield and apply imaging condition 
  call propagate_receiversdecomp(imageC,imageUUDD,imageUDDU,source_array, souUp, souDn)

  ! (4) Write wavefield movies header 
  if (wfldout) then
       call write_movie_header(wfldimg_)
       call write_movie_header(wfldsou_)
       call write_movie_header(wfldrec_)
   end if

  deallocate(source_array,souUp,souDn)
  call clean_up_array()

end subroutine  
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! SUBROUTINE #14: MODEL A SHOT
!! (1) Allocation
!! (2) Propagate and save source wavefield
!! (3) Collect data
!! (4) Output source movie (if optioned)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

subroutine model_shot(dat) 

  integer              :: it, irx
  real, dimension(:,:) :: dat !Shot gather
  real, allocatable, dimension(:,:,:) :: source_array
  complex, allocatable, dimension(:,:,:) :: souUp, souDn
  character(len=128)                  :: wfldup_, wflddn_

  write(0,*) 'Running: model_shot'
  
  ! (1) Allocation
  allocate(source_array(nz,nx,nt_data))
  allocate(souUp(nz,nx,nt_data))
  allocate(souDn(nz,nx,nt_data))

  ! (2) Propagate and save source wavefield, kz decomposition
  call create_source_array(source_array,souUp,souDn)


  ! (3) collect data
  do irx=1,size(dat,2)
     dat(:,irx) = source_array(rec_loc(1,irx),rec_loc(2,irx),:)  
  end do

  ! (4) Output source movie
  write(0,*) '@@', wfldout
  if (wfldout) then
       write(0,*) 'wfldsou_', wfldsou_, wfldup_, wflddn_
       call getwfldname(wfldsou_,wfldup_,wflddn_)
       call sep_write(source_array,wfldsou_)
       call write_movie_header(wfldsou_)
       call sep_write(souUp,wfldup_)
       call write_movie_header(wfldup_)
       call sep_write(souDn,wflddn_)
       call write_movie_header(wflddn_)

  end if

  deallocate(source_array) 
  call clean_up_array()

end subroutine  

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! SUBROUTINE #15: BORN Model
!! (1) Allocation
!! (2) Propagate and save source wavefield
!! (3) Collect data
!! (4) Output source movie (if optioned)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

subroutine born_modeldecomp(image,dat) 

  integer              :: it, irx
  real, dimension(:,:) :: image, dat
  real, allocatable, dimension(:,:,:) :: source_array
  complex, allocatable, dimension(:,:,:) :: souUp, souDn

  write(0,*) 'Running: model_shot'
  

  ! (1) Allocation
  allocate(source_array(nz,nx,nt_data))
  allocate(souUp(nz,nx,nt_data))
  allocate(souDn(nz,nx,nt_data))

  ! (2) Propagate and save source wavefield
  call create_source_array(source_array,souUp,souDn)

  ! (3) Calculate receiver wavefield and inject born source
  call propagate_borndecomp(image,dat,source_array,souUp,souDn)


  ! (4) Write wavefield movies header 
  if (wfldout) then
       call write_movie_header(wfldsou_)
       call write_movie_header(wfldrec_)
       call write_movie_header(wfldimg_)
   end if

  call clean_up_array()

end subroutine  


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
!! SUBROUTINE #16: CLEANUP! 
!! Deallocate memory specific to a shot
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 

 subroutine clean_up_array()
     deallocate (src_loc,rec_loc,src_func,rec_func) 	
 end subroutine
 
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
!! SUBROUTINE #17: Write wavefield movie header file
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
  subroutine write_movie_header(filename)

    character(len=128)  :: filename

    call auxputch("n1",'i',nz,filename)
    call auxputch("n2",'i',nx,filename)
    call auxputch("n3",'i',nt_data,filename)
    call auxputch("d1",'f',dvel(1),filename)
    call auxputch("d2",'f',dvel(2),filename)
    call auxputch("d3",'f',dt*jt,filename)
    call auxputch("o1",'f',ovel(1),filename)
    call auxputch("o2",'f',ovel(2),filename)
    call auxputch("o3",'f',0.,filename)
    call auxputch("label1",'s','z(m)',filename) 
    call auxputch("label2",'s','x(m)',filename) 
    call auxputch("label3",'s','t(s)',filename) 

   end subroutine


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! SUBROUTINE # 18: Apply tapering
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  subroutine taperingCos(prev,cur,next)

    integer  :: iz, ix
    real, pointer, dimension(:,:) :: prev,cur,next !Previous,current, and next wavefields

       ! Corner
       do ix=1,npad
          do iz=1,npad

              prev(iz,ix) = prev(iz,ix)*ctaper(iz,ix);                   ! top left
              prev(iz,nx-ix+1) = prev(iz,nx-ix+1)*ctaper(iz,ix)          ! top right
              prev(nz-iz+1,nx-ix+1) = prev(nz-iz+1,nx-ix+1)*ctaper(iz,ix)   ! bottom right
              prev(nz-iz+1,ix) = prev(nz-iz+1,ix)*ctaper(iz,ix)          ! bottom left

              cur(iz,ix) = cur(iz,ix)*ctaper(iz,ix);                   ! top left
              cur(iz,nx-ix+1) = cur(iz,nx-ix+1)*ctaper(iz,ix)          ! top right
              cur(nz-iz+1,nx-ix+1) = cur(nz-iz+1,nx-ix+1)*ctaper(iz,ix)   ! bottom right
              cur(nz-iz+1,ix) = cur(nz-iz+1,ix)*ctaper(iz,ix)          ! bottom left
              next(iz,ix) = next(iz,ix)*ctaper(iz,ix);                   ! top left
              next(iz,nx-ix+1) = next(iz,nx-ix+1)*ctaper(iz,ix)          ! top right
              next(nz-iz+1,nx-ix+1) = next(nz-iz+1,nx-ix+1)*ctaper(iz,ix)   ! bottom right
              next(nz-iz+1,ix) = next(nz-iz+1,ix)*ctaper(iz,ix)          ! bottom left
          end do
       end do

       do iz=npad+1,nz-npad
          prev(iz, 1:npad) = prev(iz, 1:npad) * zxtaper(1:npad:+1); 
          prev(iz,nx-npad+1:nx) = prev(iz,nx-npad+1:nx) * zxtaper(npad:1:-1)
          cur(iz, 1:npad) = cur(iz, 1:npad) * zxtaper(1:npad:+1);
          cur(iz,nx-npad+1:nx) = cur(iz,nx-npad+1:nx) * zxtaper(npad:1:-1)
          next(iz, 1:npad) = next(iz, 1:npad) * zxtaper(1:npad:+1);
          next(iz,nx-npad+1:nx) = next(iz,nx-npad+1:nx) * zxtaper(npad:1:-1)

       end do

       do ix=npad+1,nx-npad
          prev(1:npad,ix) = prev(1:npad,ix) * zxtaper(1:npad:+1);
          prev(nz-npad+1:nz,ix) = prev(nz-npad+1:nz,ix) * zxtaper(npad:1:-1)
          cur(1:npad,ix) = cur(1:npad,ix) * zxtaper(1:npad:+1);
          cur(nz-npad+1:nz,ix) = cur(nz-npad+1:nz,ix) * zxtaper(npad:1:-1)
          next(1:npad,ix) = next(1:npad,ix) * zxtaper(1:npad:+1);
          next(nz-npad+1:nz,ix) = next(nz-npad+1:nz,ix) * zxtaper(npad:1:-1)
       end do



  end subroutine

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$
!! SUBROUTINE # 19: Tapering Setup
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$
    subroutine init_taperCos(npad)

    integer                     :: iz, ix,ii, npad
    real                       :: dist
    real                        :: pi=3.14159265359


    allocate(zxtaper(npad), ctaper(npad,npad))

       ! all four edges
       do ii=1,npad
          zxtaper(ii) = 0.985 + 0.015*cos(pi*(npad-ii)/npad)
       end do

       ! Corner
       do ix=1,npad
          do iz=1,npad
              dist = sqrt(1.001*(npad-ix)*(npad-ix)/npad/npad + 0.999999*(npad-iz)*(npad-iz)/npad/npad)
              ctaper(iz,ix) =  0.985 + 0.015*cos(pi*dist)
          end do
       end do

  end subroutine

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$
!! SUBROUTINE # 20: Update sponge
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$
   subroutine update_sponge(prev,cur,next)

    real, pointer, dimension(:,:) :: prev,cur,next !Previous,current, and next wavefields
    real          :: dx, dz
    integer       :: ix, iz


    dz=dvel(1); dx = dvel(2);


    ! bottom sponge
    do ix=nop+1,nx-nop
       do iz=nz-npad-nop,nz-nop
          next(iz,ix) = next(iz,ix) - mult(iz,ix)*(next(iz,ix)-cur(iz,ix)+v(nz-nop,ix)*dt/dz*(cur(iz+1,ix)-cur(iz,ix)))
       end do
    end do  
    
    ! top sponge
    do ix=nop+1,nx-nop
       do iz=1+nop,npad+nop
          next(iz,ix) = next(iz,ix) - mult(iz,ix)*(next(iz,ix)-cur(iz,ix)+v(nop+1,ix)*dt/dz*(cur(iz-1,ix)-cur(iz,ix)))
       end do
    end do  

   ! right sponge
   do ix=nx-npad-nop,nx-nop
      do iz=1+nop,nz-nop
          next(iz,ix) = next(iz,ix) - mult(iz,ix)*(next(iz,ix)-cur(iz,ix)+v(iz,nx-nop)*dt/dx*(cur(iz,ix+1)-cur(iz,ix)))
      end do
   end do

  ! left sponge
   do ix=1+nop,npad+nop
      do iz=1+nop,nz-nop
          next(iz,ix) = next(iz,ix) - mult(iz,ix)*(next(iz,ix)-cur(iz,ix)+v(iz,nop+1)*dt/dx*(cur(iz,ix-1)-cur(iz,ix)))
      end do
   end do



   end subroutine

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$
!! SUBROUTINE # 21: Update the 1-way wave equation regions
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$
   subroutine update_oneway(prev,cur,next)

    real, pointer, dimension(:,:) :: prev,cur,next !Previous,current, and next wavefields
    real          :: dx, dz, delp
    integer       :: iz, ix

    dz=dvel(1); dx = dvel(2);

           ! Update the bottom operator pad
           do iz=nz-nop+1,nz
              do ix=1+nop,nx-nop
                delp=1.2*v(nz-nop,ix)*dt/dz
                cur(iz,ix)=1./(1.+delp)*((1.-delp)*(prev(iz,ix)-cur(iz-1,ix)))+prev(iz-1,ix)
              end do
           end do
            do iz=nz-nop+1,nz
               do ix=1+nop,nx-nop
                  prev(iz,ix)=cur(iz,ix)
               end do
            end do

           ! Update the top operator pad
           do iz=nop,1,-1
              do ix=1+nop,nx-nop
                delp=1.2*v(1+nop,ix)*dt/dz
                cur(iz,ix)=1./(1.+delp)*((1.-delp)*(prev(iz,ix)-cur(iz+1,ix)))+prev(iz+1,ix)
              end do
           end do
            do iz=nop,1,-1
               do ix=1+nop,nx-nop
                  prev(iz,ix)=cur(iz,ix)
               end do
            end do

           ! Update the right side operator pad
           do ix=nx-nop+1,nx
              do iz=1+nop,nz-nop
                 delp=1.2*v(iz,nx-nop)*dt/dx
                 cur(iz,ix)=1./(1.+delp)*((1.-delp)*(prev(iz,ix)-cur(iz,ix-1)))+ prev(iz,ix-1)
              end do
           end do
           do ix=nx-nop+1,nx
               do iz=1+nop,nz-nop
                  prev(iz,ix)=cur(iz,ix)
               end do
           end do
 
           ! Update the left side operator pad 
           do ix=nop,1,-1
              do iz=1+nop,nz-nop
                 delp=1.2*v(iz,1+nop)*dt/dx
                 cur(iz,ix)=1./(1.+delp)*((1.-delp)*(prev(iz,ix)-cur(iz,ix+1)))+prev(iz,ix+1)
              end do
           end do
           do ix=nop,1,-1
               do iz=1+nop,nz-nop
                 prev(iz,ix)=cur(iz,ix)
               end do
           end do
                      
   end subroutine


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$
!! SUBROUTINE # 22: Define sponge coefficients
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$
  subroutine  define_mult()
  
    integer    :: ix, iz
    real       :: dx, dz
   
   
    dx=dvel(2); dz=dvel(1);
   
    allocate(mult(nz,nx))

    mult=0.

    ! right sponge coeffs
    do ix=nx-nop-npad,nx-nop
       do iz=1+nop,nz-nop
          mult(iz,ix)=float(ix-(nx-nop-npad)+1)/float(npad+nop)*.05*((v(iz,ix)*dt/dx)/.333)
       end do
    end do
 
    ! left sponge coeffs
    do ix=1+nop,npad+nop
       do iz=1+nop,nz-nop
          mult(iz,ix)=float((npad+nop-ix)+1)/float(npad+nop)*.05*((v(iz,ix)*dt/dx)/.333)
       end do
    end do

    ! bottom sponge coeffs
    do ix=1+nop,nx-nop
       do iz=nz-nop-npad,nz-nop
          mult(iz,ix)=float(iz-(nz-nop-npad)+1)/float(npad+nop)*.05*((v(iz,ix)*dt/dz)/.333)
       end do
    end do

    ! top sponge coeffs
    do ix=1+nop,nx-nop
       do iz=1+nop,npad+nop
          mult(iz,ix)=float((npad+nop-iz)+1)/float(npad+nop)*.05*((v(iz,ix)*dt/dz)/.333)
       end do
    end do

  end subroutine


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$
!! SUBROUTINE # 23: Take derivative for 1D array
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$

  subroutine FirstDer(adj,souIn,souOut,nt,dt)

    logical :: adj
    integer :: i, nt
    real    :: dt, rdt
    real, dimension(:) :: souIn, souOut
    real, dimension(:), allocatable :: temp

   allocate(temp(nt))
   temp(1:nt) = souIn(1:nt)
    
    souOut(1:nt) = 0.
    
    
    if (.not. adj) then
       
      do i=1,nt-1
         souOut(i+1) = souOut(i+1)+ temp(i)
         souOut(i)   = souOut(i) - temp(i)
      end do
      
    else
       
      do i=1,nt-1
         souOut(i+1) = souOut(i+1) + temp(i+1)-temp(i)
      end do
         
    end if    

    deallocate(temp)
  end subroutine 

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$
!! SUBROUTINE # 24: Take Vertical Derivative of image
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$
  subroutine Ver2ndDer(adj,m,nz,nx)
  ! Take the second derivative vertically and its adjoint

  logical                   :: adj
  integer                   :: nz, nx, ix,iz
  real, dimension(:), allocatable :: temp
  real, dimension(:,:)      :: m


  allocate(temp(nz))

    do ix=1,nx
      temp(:) = m(:,ix)
      m(:,ix) = 0.
      do iz=2,nz-1
         if (adj) then
            m(iz,ix) = m(iz,ix) + (-2.*temp(iz) + temp(iz+1) + temp(iz-1))
         else
            m(iz,ix) = m(iz,ix) + -2*temp(iz)
            m(iz,ix) = m(iz,ix) + temp(iz+1)
            m(iz,ix) = m(iz,ix) + temp(iz-1)
         end if
       end do
    end do

  end subroutine


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$
!! SUBROUTINE # 25: Add "Up" and "Dn" in the filename of base_
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$

 subroutine  getwfldname(base_,wfldup_,wflddn_)

   character(len=128)   :: base_, wfldup_, wflddn_
   integer              :: p1,d2
   character*3          :: aa

   p1 = scan(base_,".H")
   wfldup_(:p1-1) = base_(:p1-1)    
   wflddn_(:p1-1) = base_(:p1-1)    
   wfldup_(p1:) = " "
   wflddn_(p1:) = " "
   wfldup_(p1:p1+1) = "Up"
   wflddn_(p1:p1+1) = "Dn"
   wfldup_(p1+2:)=".H"
   wflddn_(p1+2:)=".H"

   write(0,*) 'getwfldname', base_, wfldup_, wflddn_

 end subroutine

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$
!! SUBROUTINE # 26: Add "UU","UD","DU","DD" in the filename of base_
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$

 subroutine  getimagename(base_,nameUUDD_,nameUDDU_)

   character(len=128)   :: base_, nameUUDD_,nameUDDU_
   integer              :: p1,d2
   character*3          :: aa

   p1 = scan(base_,".H")
   nameUUDD_(:p1-1) = base_(:p1-1)    
   nameUDDU_(:p1-1) = base_(:p1-1)      
   nameUUDD_(p1:) = " "
   nameUDDU_(p1:) = " "


   nameUUDD_(p1:p1+3) = "UUDD"
   nameUDDU_(p1:p1+3) = "UDDU"
   nameUUDD_(p1+4:)=".H"
   nameUDDU_(p1+4:)=".H"


   write(0,*) 'getimagename', base_
   write(0,*) '==========', nameUUDD_,nameUDDU_

 end subroutine

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$
!! SUBROUTINE # 27: write header for image file
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$

 subroutine  writeimageheader(filename,n1,n2)
   
    character(len=128)   :: filename
    integer              :: n1, n2

    call auxputch("n1",'i',n1,filename)
    call auxputch("n2",'i',n2,filename)
    call auxputch("d1",'f',dvel(1),filename)
    call auxputch("d2",'f',dvel(2),filename)
    call auxputch("o1",'f',ovel(1),filename)
    call auxputch("o2",'f',ovel(2),filename)
    call auxputch("label1",'s','z(m)',filename) 
    call auxputch("label2",'s','x(m)',filename) 

  end subroutine

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$
!! SUBROUTINE # 28: 1D  interpolation by FFT
!! input series x, output series y, jt=dtx/dty  > 1
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$

  subroutine  interpFFT1(x,y,jt)

  integer*8             :: plan1f, plan2f
  integer               :: n1x, n1y, n1xc, n1yc 
  integer               :: i, iy, ix, stat, flags
  real, dimension(:)    :: x, y
  complex, dimension(:), allocatable :: cx, cy, cyshift, cyout
  integer               :: jt

  if (jt < 1) call erexit('interpFFT1 error: jt<1')

  n1x=size(x,1);  n1xc = n1x/2+1
  n1y=size(y,1);  n1yc = n1y/2+1

   write(0,*) 'n1x, n1y, jt', n1x, n1y, jt
   write(0,*) 'n1xc, n1yc', n1xc,n1yc

   if (allocated(cx)) deallocate(cx)
   if (allocated(cy)) deallocate(cy)

  allocate(cx(n1xc),cy(n1yc))

!   write(0,*) 'size ', size(x), size(cx)
!   write(0,*) 'size ', size(y), size(cy)

  !initialization
   do i=1,n1xc
    cx(i)=cmplx(0.,0.);
   end do
   do i=1,n1yc
     cy(i)=cmplx(0.,0.)
   end do   

  !! Do forward FFT on x
   call sfftw_plan_dft_r2c_1d(plan1f, n1x, x, cx, FFTW_ESTIMATE)
   call sfftw_execute(plan1f)
   call sfftw_destroy_plan(plan1f)

  !! copy cx to cy and pad cy with zeros
  do ix=1,n1xc
     cy(ix) = cx(ix)
  end do 

  !! Do IFFT on cy to get y, done!
   call sfftw_plan_dft_c2r_1d(plan2f, n1y, cy, y, FFTW_ESTIMATE)
   call sfftw_execute(plan2f)
   call sfftw_destroy_plan(plan2f)

   y=y/n1x

   deallocate(cx,cy)

  end subroutine interpFFT1

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$
!! SUBROUTINE # 29: 2D  interpolation by FFT
!! input series x, output series y, jt=dtx/dty  > 1
!! Do a series of 1D interpolation along the first axis
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$

  subroutine  interpFFT2(x,y,jt)

  integer*8             :: plan1f, plan2f
  integer               :: n1x, n1y, n1xc, n1yc 
  integer               :: n2x
  integer               :: i, iy, ix, stat, flags, i2
  real, dimension(:,:)  :: x, y
  complex, dimension(:), allocatable :: cx, cy 
  real, dimension(:), allocatable    :: out,in
  integer               :: jt

  if (jt < 1) call erexit('interpFFT2 error: jt<1')

  n1x=size(x,1);  n1xc = n1x/2+1
  n1y=size(y,1);  n1yc = n1y/2+1
  n2x=size(x,2);

   write(0,*) 'n1x, n1y, jt', n1x, n1y, jt
   write(0,*) 'n1xc, n1yc', n1xc,n1yc

   if (allocated(cx)) deallocate(cx)
   if (allocated(cy)) deallocate(cy)
   if (allocated(in)) deallocate(in)
   if (allocated(out)) deallocate(out)

  allocate(cx(n1xc),cy(n1yc),in(n1x),out(n1y))

!   write(0,*) 'size ', size(x), size(cx)
!   write(0,*) 'size ', size(y), size(cy)

  !initialization
   do i=1,n1xc
    cx(i)=cmplx(0.,0.);
   end do
   do i=1,n1yc
     cy(i)=cmplx(0.,0.)
   end do  
   

  !! Do forward FFT on x
   call sfftw_plan_dft_r2c_1d(plan1f, n1x, in, cx, FFTW_ESTIMATE)
  !! Do IFFT on cy to get y, done!
   call sfftw_plan_dft_c2r_1d(plan2f, n1y, cy, out, FFTW_ESTIMATE)
   do i2=1,n2x
      cx=cmplx(0.,0.)
      cy=cmplx(0.,0.)
      in(1:n1x) = x(1:n1x,i2)
      call sfftw_execute(plan1f)
      !! copy cx to cy and pad cy with zeros
      do ix=1,n1xc
      cy(ix) = cx(ix)
      end do 
      call sfftw_execute(plan2f)
      y(1:n1y,i2)=out(1:n1y)
   end do
   call sfftw_destroy_plan(plan1f)
   call sfftw_destroy_plan(plan2f)

   y=y/n1x

   deallocate(cx,cy,in,out)

  end subroutine interpFFT2



 end module

 


